MODULE mo_example

  USE error_handling
  USE mtime
  IMPLICIT NONE

CONTAINS

  INTEGER FUNCTION division(number1, number2, ifail)
    INTEGER,       INTENT(IN)    :: number1, number2
    TYPE(t_error), INTENT(INOUT) :: ifail

    division = 0
    IF (number2 == 0) THEN
      CALL create_error(ifail, t_arithmetic_error("Division by zero!"))
    ELSE IF ((number1 < 0) .OR. (number2 < 0)) THEN
      CALL create_error(ifail, t_invalid_input_error("Only positive numbers allowed!"))
    ELSE
      division = number1 / number2
    END IF
  END FUNCTION division

  SUBROUTINE event_tests(ref_date,start_date,end_date,intvl,query_date)
    CHARACTER(LEN=*), intent(IN) :: ref_date,start_date,end_date,intvl,query_date

    type(event), pointer :: outputEvent
    TYPE(datetime), POINTER :: current_date_test, dtt
    type(timedelta), pointer :: tdd
    character(len=max_datetime_str_len)  :: current_date_string_tmp


    WRITE (0,*) "EXAMPLES FOR EVENT HANDLING"
    outputEvent => newEvent('test', ref_date,start_date,end_date,intvl)

    dtt => getEventReferenceDateTime(outputEvent) 
    call datetimeToString(dtt, current_date_string_tmp)
    print *, trim(current_date_string_tmp)

    dtt => getEventFirstDateTime(outputEvent)
    call datetimeToString(dtt, current_date_string_tmp)
    print *, trim(current_date_string_tmp)

    dtt => getEventLastDateTime(outputEvent)
    call datetimeToString(dtt, current_date_string_tmp)
    print *, trim(current_date_string_tmp)

    tdd => getEventInterval(outputEvent)
    call timedeltaToString(tdd, current_date_string_tmp)
    print *, trim(current_date_string_tmp)

    current_date_test => newDatetime(query_date)
  
    PRINT *, isCurrentEventActive(outputEvent, current_date_test)

    call deallocateEvent(outputEvent)
    call deallocateDatetime(current_date_test)
    
   end subroutine event_tests

END MODULE mo_example

PROGRAM example
  
  USE mo_example
  
  USE mtime,          ONLY: setCalendar, PROLEPTIC_GREGORIAN  
  USE mtime_hl,       ONLY: t_datetime, t_timedelta, to_string 
  USE mtime_hl,       ONLY: to_string, assignment(=), operator(+), operator(-), operator(==), &
          & operator(/=), operator(<), operator(>), operator(<=), operator(>=)
  USE error_handling
  
  IMPLICIT NONE
  
  TYPE(t_datetime)  :: dt, dt2, dt3
  TYPE(t_timedelta) :: td
  
  INTEGER       :: test_number1, test_number2, test_result
  TYPE(t_error) :: ifail

  WRITE (0,*) "example_hl : test example"

  WRITE (0,*) "  setting calendar."
  CALL setCalendar(PROLEPTIC_GREGORIAN)

  ! Instead of:  TYPE(datetime), POINTER :: dt
  !              dt => newDatetime("1970-01-01T00:00:00")
  !                [...]
  !              CALL deallocateDatetime(dt)

  ! TYPE(t_datetime)  :: dt
  WRITE (0,*) "  testing assignment of t_datetime."
  dt = t_datetime("1970-01-01T00:00:00")

  ! Instead of:  TYPE(timedelta), POINTER :: td
  !              td => newTimeDelta("PT1H1M1S")
  !                [...]
  !              CALL deallocateTimedelta(td)

  ! TYPE(t_timedelta) :: td
  WRITE (0,*) "  testing assignment of t_timedelta."
  td = t_timedelta("PT1H1M1S")

  ! Instead of:  CALL datetimeToString(dt, dstring1)
  !              CALL timedeltaToString(td, dstring2)
  !              WRITE (0,*) "compute ", dstring1, " + ", dstring2

  WRITE (0,*) "compute ", to_string(dt), " + ", to_string(td)

  ! Adding time intervals is unchanged:

  dt = dt + td
  WRITE (0,*) "result: ", to_string(dt)

  ! --- Further examples:

  ! subtraction of two dates
  dt2 = t_datetime("1970-01-01T00:00:00")
  td = dt-dt2
  WRITE (0,*) "subtraction of dates: time delta: ", to_string(td)

  ! comparison of dates
  dt  = t_datetime("1970-01-01T00:00:00")
  dt2 = t_datetime("1970-01-01T00:00:00")
  WRITE (0,*) to_string(dt), " == ", to_string(dt2), ": ", (dt == dt2)
  dt3 = t_datetime("1970-01-01T00:00:01")
  WRITE (0,*) to_string(dt), " == ", to_string(dt3), ": ", (dt == dt3)

  ! interval assignment with milliseconds
  td = t_timedelta(360000)
  WRITE (0,*) "interval assignment with milliseconds: ", to_string(td)

  ! ------------------------------------------------------------
  ! EXAMPLE FOR ERROR HANDLING
  ! ------------------------------------------------------------

  WRITE (0,*) "EXAMPLE FOR ERROR HANDLING"
  test_number1 = 10
  test_number2 = -1
  test_result  = division(test_number1, test_number2, ifail)

  SELECT TYPE(info => ifail%info)
  TYPE is (t_arithmetic_error)
    WRITE (0,*) "I'm too stupid!"
    CALL ifail%report()
  TYPE is (t_invalid_input_error)
    WRITE (0,*) "You are too stupid!"
    CALL ifail%report()
  CLASS default
    WRITE (0,*) "Everything was ok!"
    WRITE (0,*) "result = ", test_result
  END SELECT


  ! ------------------------------------------------------------
  ! EXAMPLES FOR EVENT HANDLING
  ! ------------------------------------------------------------

  ! expected result: T
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2010-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'PT06H', query_date='2010-01-02T00:00:00')

  ! expected result: F
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2010-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'P1M', query_date='2010-01-02T00:00:00')

  ! expected result: F
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2010-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'P06Y', query_date='2010-01-02T00:00:00')

  ! expected result: T
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2001-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'P06Y', query_date='2006-01-01T00:00:00')

  ! expected result: F
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2007-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'P06Y', query_date='2006-01-01T00:00:00')

   ! expected result: T
  CALL event_tests('2000-01-01T00:00:00', &
      &           '2007-01-01T00:00:01', &
      &           '2013-01-01T00:00:02', &
      &           'P06Y', query_date='2012-01-01T00:00:00')
END PROGRAM example
