! TODO: 1) initializer with generic t_datetime and t_timedelta, extent for t_julianday and t_juliandelta
!       2) No type-bound procedures (Fortran standard problem) but generics (to_string)
!       3) Operator/assignment as generics (=,+,-) instead of type-bound procedures
!       4) Additional functions: to_days, to_hours, to_seconds
!          - result is t_juliandelta internal
!          - external interface argument interface:
!            to_days, to_hours, to_seconds; arguments:
!            (t_timdelta, anchor = t_datetime, error_code 
!                in        in, optional        out, optional
!            return value type (val (int), rem (t_timedelta))
!            if error_code present user handles problem
!          - on error (because of missing anchor), kill the cat (?!) 
!       5a) Division: no overloaded operator, instead: subroutine. return type: t_divison_result (aka. t_conversion_result) 
!       5) Overload REAL operator for division result and types from 4)
!       6) create cache object for struct _event: all short-cuts to special evaluation, e.g. 
!          eventisFirstInDay should move there
!       7) create POINTER-less high level interface for events
!       8) repair event trigger algorithm (years in bisection)
!    
!
!> @file libmtime.f90
!!
!! @brief Providing a high level interface for the Fortran language bindings of libmtime
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Florian Prill, DWD
!!
!! @defgroup FortranBindings libmtime high level Fortran language bindings
!! @{
!!
module mtime_hl

  use mtime
  use error_handling

  implicit none
  
  private

  public :: t_datetime, t_timedelta
  public :: to_string, assignment(=), operator(+), operator(-), operator(==), &
          & operator(/=), operator(<), operator(>), operator(<=), operator(>=)
  integer, parameter :: i8 = selected_int_kind(14)   !< at least 8 byte integer

  !> Wrapper class for "mtime" data type "datetime".
  !
  type t_datetime
    private
    type(datetime) :: dt
  end type t_datetime

  !> Wrapper class for "mtime" data type "timedelta".
  !
  type t_timedelta
    private
    type(timedelta) :: td
  end type t_timedelta

  interface assignment(=)
    module procedure :: assign_t_datetime
    module procedure :: assign_t_timedelta
  end interface assignment(=)

  interface operator(+)
    module procedure :: t_datetime_add_timedelta
  end interface operator(+)

  interface operator(-)
    module procedure :: t_datetime_sub_timedelta
    module procedure :: t_datetime_sub_datetime
  end interface operator(-)

  interface operator(==)
    module procedure :: t_datetime_equal
  end interface operator(==)

  interface operator(/=)
    module procedure :: t_datetime_not_equal
  end interface operator(/=)

  interface operator(<)
    module procedure :: t_datetime_less_than
  end interface operator(<)

  interface operator(>)
    module procedure :: t_datetime_greater_than
  end interface operator(>)

  interface operator(<=)
    procedure :: t_datetime_less_or_equal
  end interface operator(<=)

  interface operator(>=)
    procedure :: t_datetime_greater_or_equal
  end interface operator(>=)

  interface to_string
    module procedure :: t_datetime_to_string
    module procedure :: t_timedelta_to_string
  end interface to_string

  interface t_datetime
    module procedure t_datetime_assign_string
  end interface t_datetime

  interface t_timedelta
    module procedure t_timedelta_assign_string
    module procedure t_timedelta_assign_ms
  end interface t_timedelta

  !> Error info type: mtime error
  type, extends(t_error_info) :: t_mtime_error
  end type t_mtime_error

contains

  !___________________________________________________________________________
  ! datetime section:
  !
  ! generic assignment for constructors
  !
  subroutine assign_t_datetime(to, from)
    class(t_datetime), intent(out) :: to
    class(t_datetime), intent(in) :: from
    type(datetime), pointer :: dt_tmp
    dt_tmp => newdatetime(from%dt)
    to%dt = dt_tmp
    call deallocatedatetime(dt_tmp)
  end subroutine assign_t_datetime

  ! constructor for a datetime string
  !
  type(t_datetime) function t_datetime_assign_string(dt_string)
    character(len=*), intent(in) :: dt_string
    type(datetime), pointer :: dt_tmp
    dt_tmp => newdatetime(dt_string)
    t_datetime_assign_string%dt = dt_tmp
    call deallocatedatetime(dt_tmp)
  end function t_datetime_assign_string

  ! Convert t_datetime object to string.
  !
  function t_datetime_to_string(this, opt_format_string, opt_error)
    character(len=max_datetime_str_len)     :: t_datetime_to_string
    type(t_datetime)                        :: this
    character(len=*), intent(in),  optional :: opt_format_string
    type(t_error),    intent(out), optional :: opt_error

    type(datetime), pointer                 :: dt_tmp
    integer                                 :: errno
    character(len=max_error_str_len)        :: error_str

    dt_tmp => newdatetime(this%dt)
    if (present(opt_format_string)) then
      call datetimetoposixstring(dt_tmp, t_datetime_to_string, opt_format_string, errno)
    else
      call datetimetostring(dt_tmp, t_datetime_to_string, errno)
    endif

    if (present(opt_error)) then
      if  (errno /= 0) then
        call mtime_strerror(errno, error_str)
        call create_error(opt_error, t_mtime_error(error_str))
      end if
    end if
    call deallocatedatetime(dt_tmp)
  end function t_datetime_to_string

  ! Convert t_timedelta object to string.
  !
  function t_timedelta_to_string(this, opt_error)
    character(len=max_timedelta_str_len) :: t_timedelta_to_string
    type(t_timedelta)                    :: this
    type(t_error), intent(out), optional :: opt_error

    type(timedelta), pointer             :: td_tmp
    INTEGER                              :: errno
    character(len=max_error_str_len)     :: error_str

    td_tmp => newtimedelta(this%td)
    call timedeltatostring(td_tmp, t_timedelta_to_string, errno)

    if (present(opt_error)) then
      if  (errno /= 0) then
        call mtime_strerror(errno, error_str)
        call create_error(opt_error, t_mtime_error(error_str))
      end if
    end if
    call deallocatetimedelta(td_tmp)
  end function t_timedelta_to_string

  ! Addition of time interval to datetime object.
  !
  function t_datetime_add_timedelta(this, td) result(dt_td_sum)
    type (t_datetime)               :: dt_td_sum
    class (t_datetime),  intent(in) :: this    
    class (t_timedelta), intent(in) :: td
    type(datetime),  pointer :: dt_tmp
    type(timedelta), pointer :: td_tmp
    dt_tmp => newDatetime(this%dt)
    td_tmp => newTimedelta(td%td)
    dt_tmp = dt_tmp + td_tmp
    dt_td_sum%dt = dt_tmp
    call deallocateDatetime(dt_tmp)
    call deallocateTimedelta(td_tmp)
  end function t_datetime_add_timedelta

  ! Subtraction of time interval to datetime object.
  !
  function t_datetime_sub_timedelta(this, td) result(dt_td_sum)
    type (t_datetime)               :: dt_td_sum
    class (t_datetime),  intent(in) :: this    
    class (t_timedelta), intent(in) :: td
    type(datetime),  pointer :: dt_tmp
    type(timedelta), pointer :: td_tmp
    dt_tmp => newDatetime(this%dt)
    td_tmp => newTimedelta(td%td)
    if (td_tmp%sign == "+") then
      td_tmp%sign = "-" 
    else
      td_tmp%sign = "+"       
    endif
    dt_tmp = dt_tmp + td_tmp
    dt_td_sum%dt = dt_tmp
    call deallocateDatetime(dt_tmp)
    call deallocateTimedelta(td_tmp)
  end function t_datetime_sub_timedelta

  ! Subtraction of two dates.
  !
  function t_datetime_sub_datetime(this, dt) result(dt_dt_diff)
    type (t_timedelta)              :: dt_dt_diff
    class (t_datetime),  intent(in) :: this
    class (t_datetime),  intent(in) :: dt
    dt_dt_diff%td = getTimeDeltaFromDateTime(this%dt, dt%dt)
  end function t_datetime_sub_datetime

  ! Overloaded operator: test for equivalence.
  !
  logical function t_datetime_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_equal = (this%dt == dt%dt)
  end function t_datetime_equal

  logical function t_datetime_not_equal(this, dt)
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_not_equal = (this%dt /= dt%dt)
  end function t_datetime_not_equal

  logical function t_datetime_less_than(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_less_than = (this%dt < dt%dt)
  end function t_datetime_less_than

  logical function t_datetime_greater_than(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_greater_than = (this%dt > dt%dt)
  end function t_datetime_greater_than

  logical function t_datetime_less_or_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_less_or_equal = (this%dt <= dt%dt)
  end function t_datetime_less_or_equal

  logical function t_datetime_greater_or_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_greater_or_equal = (this%dt >= dt%dt)
  end function t_datetime_greater_or_equal

  !___________________________________________________________________________
  ! timedelta section:
  !
  ! generic assignment for constructors
  !
  subroutine assign_t_timedelta(to, from)
    class(t_timedelta), intent(out) :: to
    class(t_timedelta), intent(in) :: from
    type(timedelta), pointer :: td_tmp
    td_tmp => newTimedelta(from%td)
    to%td = td_tmp
    to%td%sign = td_tmp%sign
    call deallocateTimedelta(td_tmp)
  end subroutine assign_t_timedelta

  ! constructor for timedelta string
  !
  type(t_timedelta) function t_timedelta_assign_string(td_string)
    character(len=*), intent(in)  :: td_string
    type(timedelta), pointer :: td_tmp
    td_tmp => newtimedelta(td_string)
    t_timedelta_assign_string%td = td_tmp
    t_timedelta_assign_string%td%sign = td_tmp%sign
    call deallocatetimedelta(td_tmp)
  end function t_timedelta_assign_string

  ! constructor for integer milliseconds (integer)
  !
  type(t_timedelta) function t_timedelta_assign_ms(td_ms)
    integer, intent(in) :: td_ms
    type(timedelta), pointer :: td_tmp
    character(len=max_timedelta_str_len) :: td_string
    call getptstringfromms(int(td_ms,i8), td_string)
    td_tmp => newtimedelta(td_string)
    t_timedelta_assign_ms%td = td_tmp
    t_timedelta_assign_ms%td%sign = td_tmp%sign
    call deallocatetimedelta(td_tmp)
  end function t_timedelta_assign_ms


end module mtime_hl
!>
!! @}
