/**
 * @file mtime_eventHandling.c
 *
 * @brief Event-groups which contains a list of events.
 *
 * @author Luis Kornblueh, Rahul Sinha. MPIM.
 * @date March 2013
 *
 * @note
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include "mtime_eventHandling.h"

#include "mtime_datetime.h"
#include "mtime_timedelta.h"
#include "mtime_julianDay.h"
#include "mtime_eventList.h"
#include "mtime_iso8601.h"

/* Local static functions. */
static struct _datetime*
getEventFirstTriggerDateTime(struct _datetime*, struct _timedelta*, struct _timedelta*, struct _datetime*, struct _datetime*);

// The IDs are unique only when they are live. Once a group/event has been deleted, the IDs will be reused.
// Currently, the IDs are generated but not used.
static int64_t EVENTGROUPID = 0;
static int64_t EVENTID = 0;


/**
 * @brief Construct new event-Group using a string.
 *
 * @param  egn
 *         A pointer to char. This string contains the name of event group.
 *
 * @return eg
 *         A pointer to a initialized event-Group. 
 *
 */

struct _eventGroup*
newEventGroup(const char* egn)
{
if ((egn != NULL) && (getCalendarType()))
{
  struct _eventGroup* eg = (struct _eventGroup*)calloc(1,sizeof(struct _eventGroup));
  if (eg == NULL)
    return NULL ;

  /* Copy Group name. */
  eg->eventGroupName = (char*)calloc(MAX_GROUPNAME_STR_LEN,sizeof(char));
  if (eg->eventGroupName == NULL )
    {
      free(eg);
      eg = NULL;
      return NULL ;
    }
  strncpy(eg->eventGroupName, egn, MAX_GROUPNAME_STR_LEN-1);
  eg->eventGroupName[MAX_GROUPNAME_STR_LEN - 1] = '\0';

  /* Generate eventGroupId. */
  EVENTGROUPID = EVENTGROUPID + 1;
  eg->eventGroupId = EVENTGROUPID;

  /* Initialize a NULL pointer. Stores a List of events associated with 'this' Event Group. */
  eg->rootEvent = NULL;

  return eg;
}
else
  return NULL;
}


/**
 * @brief Destructor of EventGroup.
 *
 * @param  eg
 *         A pointer to struct _eventGroup. eg is deallocated.
 */

void
deallocateEventGroup(struct _eventGroup* eg)
{
  if (eg != NULL )
    {
      EVENTGROUPID = EVENTGROUPID - 1;

      if(eg->eventGroupName)
        {
          free(eg->eventGroupName);
          eg->eventGroupName = NULL;
        }

      /* Deallocate all events in the list. */
      deallocateEventsInGroup(eg);

      free(eg);
      eg = NULL;
    }
}


/**
 * @brief Add new event to an eventgroup.
 *
 * @param  e
 *         A pointer to struct _event. The event to be added.
 *
 * @param  eg
 *         A pointer to struct _eventGroup. The eventGroup where the event is added.
 *
 * @return bool
 *         true/false indicating success or failure of addition.
 */

bool
addNewEventToEventGroup(struct _event* e, struct _eventGroup* eg)
{
  if ((e != NULL) && (eg != NULL))
    return addNewEventToGroup(e,eg);
  else
    return false;
}


/**
 * @brief Remove event from eventgroup. CRITICAL: Also, deallocate the event.
 *
 * @param  en
 *         A pointer to char. The name of event to be removed.
 *
 * @param  eg
 *         A pointer to struct _eventGroup. The eventGroup to which this event belongs.
 *
 * @return bool
 *         true/false indicating success or failure of removal.
 */

bool
removeEventFromEventGroup(char* en, struct _eventGroup* eg)
{
  if ((en != NULL )&& (eg != NULL) )
    return removeEventWithNameFromGroup(en,eg);
  else
    return false;
}

int64_t
getEventGroupId(struct _eventGroup* eg)
{
  if (eg != NULL)
    return eg->eventGroupId;
  else
    return 0;
}

char*
getEventGroupName(struct _eventGroup* eg, char* gname)
{
  if ((eg != NULL) && (gname != NULL))
    {
      strncpy(gname, eg->eventGroupName, MAX_GROUPNAME_STR_LEN-1);
      gname[MAX_GROUPNAME_STR_LEN-1] = '\0';
      return gname;
    }
  else
    return NULL;
}

struct _event*
getEventGroupRootEvent(struct _eventGroup* eg)
{
  if (eg != NULL)
    return eg->rootEvent;
  else
    return NULL;
}


  /**
   * @brief Construct new event using strings.
   *
   * To initialize an event, the event name and event interval must be specified. The reference date 
   * (also known as the anchor-date), start date and event last date are optional. If first date is not
   * specified, first date is initialized as 0-01-01T00:00:00.000. If reference date is not specified, first date 
   * is copied to reference date. If event last date is not defined, it is left as NULL and is logically equivalent 
   * to Infinity.
   *
   * Notice that events trigger every _eventReferenceDT + Integral multiple of _eventInterval strictly in [_eventFirstDT,_eventLastDT].
   *
   * @param  _en
   *         A pointer to char. This string contains the name of event. String name must be defined.
   *
   * @param  _eventReferenceDT
   *         A pointer to char. This string contains the Anchor date. Events are triggered at Anchor date + (N * eventInterval) where N is in Z.
   *         This can be NULL in which case Anchor Date = First Date.
   *
   * @param  _eventFirstDT
   *         A pointer to char. This string contains the Starting datetime (A) of the DateTime range [A-B] in which trigger is allowed. 
   *         Can be NULL.
   *
   * @param  _eventLastDT
   *         A pointer to char. This string contains the Ending datetime (B) of the DateTime range [A-B] in which trigger is allowed.
   *         Can be NULL. If defined, events will not trigger beyond this point. Else, equivalent to infinity.
   *
   * @param  _eventOffset
   *         A pointer to char. Adds a logical shift to _eventReferenceDT and the shifted value is used as the anchor date.
   *         Can be NULL; Logically equivalent to 0 shift.
   *
   * @param  _eventInterval
   *         A pointer to char. This string contains the timestep. This must be defined for every event.
   *
   * @return e
   *         A pointer to an initialized event. 
   *
   */

struct _event*
newEvent(const char* _en, 
	 const char* _eventReferenceDT, 
	 const char* _eventFirstDT, 
	 const char* _eventLastDT, 
	 const char* _eventInterval, 
	 const char* _eventOffset)
{
  if ((_en != NULL) &&  (_eventInterval != NULL) && getCalendarType())
    {
      struct _event* e = (struct _event*)calloc(1,sizeof(struct _event));
      if (e == NULL )
        return NULL;

			e->cache.cacheIsValid = false;

      /* Initialize a null pointer. Connectes to the 'next' event in the list of Events in an Event Group.  */
      e->nextEventInGroup = NULL;

      /* Copy event name. */
      e->eventName = (char*)calloc(MAX_EVENTNAME_STR_LEN,sizeof(char));
      if (e->eventName == NULL )
        goto cleanup_and_exit;
      strncpy(e->eventName, _en, MAX_EVENTNAME_STR_LEN-1);
      e->eventName[MAX_EVENTNAME_STR_LEN - 1] = '\0';

      /* Generate eventId. */
      EVENTID = EVENTID + 1;
      e->eventId = EVENTID;
      
      /* Last evaluation date */
      e->cache.eventsLastEvaluationDateTime = NULL;
  
      /* Start Date. */
      if (_eventFirstDT)
        {
          e->eventFirstDateTime = newDateTime(_eventFirstDT);
        }
      else
        e->eventFirstDateTime = newDateTime(initDummyDTString); // Default start date = "0-01-01T00:00:00.000".
  
      if (e->eventFirstDateTime == NULL )
        goto cleanup_and_exit;

      /* Anchor date. */
      if (_eventReferenceDT)
        e->eventReferenceDateTime = newDateTime(_eventReferenceDT);
      else
        e->eventReferenceDateTime = constructAndCopyDateTime(e->eventFirstDateTime); // If Anchor date is not defined, Anchor Date = First Date.
      
      if (e->eventReferenceDateTime == NULL )
        goto cleanup_and_exit;
      

      /* Last date. */
      if (_eventLastDT)
        {
          e->eventLastDateTime = newDateTime(_eventLastDT);
          if (e->eventLastDateTime == NULL )
            goto cleanup_and_exit;
        }
      else
        {
          e->eventLastDateTime = NULL; // Logically equivalent to Inf.
        } 

      /* _eventInterval must be defined. */
      e->eventInterval = newTimeDelta(_eventInterval);
      if (e->eventInterval == NULL)
        goto cleanup_and_exit;
      
      if (e->eventInterval->year == 0 && e->eventInterval->month == 0 && e->eventInterval->day == 0
          && e->eventInterval->hour == 0 && e->eventInterval->minute == 0 && e->eventInterval->second == 0 && e->eventInterval->ms == 0)
        {
          e->neverTriggerEvent   = true;
          
          e->cache.triggerCurrentEvent = false;
          
          e->cache.nextEventIsFirst    = false;
          e->cache.lastEventWasFinal   = false;
          
          /* Removed to prevent segfaults ... deallocateTimeDelta(e->eventInterval); */
          
          e->eventOffset = NULL;      
          e->cache.triggeredPreviousEventDateTime = NULL;
          e->cache.triggerNextEventDateTime = NULL;
          
          return e;
        }
      
      if (_eventOffset)
        {
          e->eventOffset = newTimeDelta(_eventOffset);
          if (e->eventOffset == NULL)
            goto cleanup_and_exit;
        }
      else
        e->eventOffset = NULL; //Logically equivalent to 0.
      
      /* Initialize with 'some' (arbitrary) value. */
      e->cache.triggeredPreviousEventDateTime = newDateTime(initDummyDTString);
      if (e->cache.triggeredPreviousEventDateTime == NULL)
        goto cleanup_and_exit;
      
      e->cache.triggerNextEventDateTime = newDateTime(initDummyDTString);
      /* triggerNextEventDateTime holds the DateTime when an event should be triggered subjectively (i.e assuming real time starts at -Inf and moves fwd.)
         At init (right now), triggerNextEventDateTime stores the DateTime of First-Ever trigger. 
         Events trigger at e->eventReferenceDateTime (+ e-> eventOffset ) + N * e->eventInterval where N is a positive, negative integer.
         Hence, first trigger happens at the nearest possible trigger >= e->eventFirstDateTime. */
      if (!getEventFirstTriggerDateTime(e->eventFirstDateTime, e->eventInterval, e->eventOffset, e->eventReferenceDateTime, e->cache.triggerNextEventDateTime))
        goto cleanup_and_exit;
      
      /* Init the Flags. */
      e->neverTriggerEvent          = false;
      
      e->cache.triggerCurrentEvent 	= false;
      
      e->cache.nextEventIsFirst 		= true;
      e->cache.lastEventWasFinal          = false;
      
      return e;
      
    cleanup_and_exit:
      
      if (e)
        {
          if (e->eventName)
            {
              free(e->eventName);
              e->eventName = NULL;
            }
          deallocateDateTime(e->eventReferenceDateTime);
          e->eventReferenceDateTime = NULL;
          deallocateDateTime(e->eventFirstDateTime);
          e->eventFirstDateTime = NULL;
          deallocateDateTime(e->eventLastDateTime);
          e->eventLastDateTime = NULL;
          deallocateTimeDelta(e->eventInterval);
          e->eventInterval = NULL;
          deallocateTimeDelta(e->eventOffset);
          e->eventOffset = NULL;
          deallocateDateTime(e->cache.triggeredPreviousEventDateTime);
          e->cache.triggeredPreviousEventDateTime = NULL;
          deallocateDateTime(e->cache.triggerNextEventDateTime);
          e->cache.triggerNextEventDateTime = NULL;
          free(e);
          e = NULL;
        }
      return NULL ;
    }
  else
    return NULL;
}


  /**
   * @brief Construct new event using data-types.
   *
   * To initialize an event, the event name and event interval must be specified. The reference date 
   * (also known as the anchor-date), start date and event last date are optional. If first date is not
   * specified, first date is initialized as 0-01-01T00:00:00.000. If reference date is not specified, first date 
   * is copied to reference date. If event last date is not defined, it is left as NULL and is logically equivalent 
   * to Infinity.
   *
   * Notice that events trigger every _eventReferenceDT + Integral multiple of _eventInterval strictly in [_eventFirstDT,_eventLastDT].
   *
   * @param  _en
   *         A pointer to char. This string contains the name of event.
   * @param  _eventReferenceDT
   *         A pointer to struct _datetime. This pointer contains the First Reference date also called anchor date. 
   *         This can be NULL, but if defined acts as the true starting datetime overriding e->eventFirstDateTime.
   * @param  _eventFirstDT
   *         A pointer to struct _datetime. This pointer contains the Starting datetime. This must be defined for every Event.
   * @param  _eventLastDT
   *         A pointer to struct _datetime. This pointer contains the Ending datetime.
   *         This can be NULL. If defined, events will not trigger beyond this point.
   * @param  _eventInterval
   *         A pointer to struct _timedelta. This pointer contains the timestep. This must be defined for every event.
   * @return e
   *         A pointer to an initialized event. 
   *
   */

struct _event*
newEventWithDataType(const char* _en, 
			struct _datetime*  _eventReferenceDT, 
			struct _datetime*  _eventFirstDT, 
			struct _datetime*  _eventLastDT, 
			struct _timedelta* _eventInterval, 
			struct _timedelta* _eventOffset)
{
  if ((_en != NULL) &&  (_eventInterval != NULL) && getCalendarType())
  {
  struct _event* e = (struct _event*)calloc(1,sizeof(struct _event));
  if (e == NULL )
    return NULL;

	e->cache.cacheIsValid = false;

  /* Initialize a null pointer. Connectes to the 'next' event in the list of Events in an Event Group.  */
  e->nextEventInGroup = NULL;

  /* Copy event name. */
  e->eventName = (char*)calloc(MAX_EVENTNAME_STR_LEN,sizeof(char));
  if (e->eventName == NULL )
    goto cleanup_and_exit;
  strncpy(e->eventName, _en, MAX_EVENTNAME_STR_LEN-1);
  e->eventName[MAX_EVENTNAME_STR_LEN - 1] = '\0';

  /* Generate eventId. */
  EVENTID = EVENTID + 1;
  e->eventId = EVENTID;

  /* Start Date. */
  if (_eventFirstDT)
    e->eventFirstDateTime = constructAndCopyDateTime(_eventFirstDT);
  else
    e->eventFirstDateTime = newDateTime(initDummyDTString); // Default start date = "0-01-01T00:00:00.000".

  if (e->eventFirstDateTime == NULL )
    goto cleanup_and_exit;


  /* Anchor date. */
  if (_eventReferenceDT)
      e->eventReferenceDateTime = constructAndCopyDateTime(_eventReferenceDT);
  else
      e->eventReferenceDateTime = constructAndCopyDateTime(e->eventFirstDateTime); // If Anchor date is not defined, Anchor Date = First Date.

   if (e->eventReferenceDateTime == NULL )
     goto cleanup_and_exit;


  /* Last date. */
  if (_eventLastDT)
      e->eventLastDateTime = constructAndCopyDateTime(_eventLastDT);
  else
      e->eventLastDateTime = NULL; // Logically equivalent to Inf.


  if (_eventInterval->year == 0 && _eventInterval->month == 0 && _eventInterval->day == 0
      && _eventInterval->hour == 0 && _eventInterval->minute == 0 && _eventInterval->second == 0 && _eventInterval->ms == 0)
    {
      e->neverTriggerEvent   = true;
      
      e->cache.triggerCurrentEvent = false;

      e->cache.nextEventIsFirst    = false;
      e->cache.lastEventWasFinal   = false;
      
      /* Added to prevent segfaults ... */
      e->eventInterval = constructAndCopyTimeDelta(_eventInterval);      
      e->eventOffset = NULL;      
      e->cache.triggeredPreviousEventDateTime = NULL;
      e->cache.triggerNextEventDateTime = NULL;

      return e;
    }
  
  /* _eventInterval must be defined. */
  e->eventInterval = constructAndCopyTimeDelta(_eventInterval);
  if (e->eventInterval == NULL)
    goto cleanup_and_exit;

  
  /* Event trigger offset. */
  if (_eventOffset)
    e->eventOffset = constructAndCopyTimeDelta(_eventOffset);
  else
    e->eventOffset = NULL;

  /* Initialize with 'some'(arbitrary) value. */
  e->cache.triggeredPreviousEventDateTime = newDateTime(initDummyDTString);
  if (e->cache.triggeredPreviousEventDateTime == NULL)
    goto cleanup_and_exit;

  e->cache.triggerNextEventDateTime = newDateTime(initDummyDTString);
  /* triggerNextEventDateTime holds the DateTime when an event should be triggered subjectively (i.e assuming real time starts at -Inf and moves fwd.)
     At init (right now), triggerNextEventDateTime stores the DateTime of First-Ever trigger. 
     Events trigger at e->eventReferenceDateTime (+ e->eventOffset) + N*e->eventInterval where N is a positive, negative or zero integer.
     Hence, first trigger happens at the nearest possible trigger >= e->eventFirstDateTime. */
  if (!getEventFirstTriggerDateTime(e->eventFirstDateTime, e->eventInterval, e->eventOffset, e->eventReferenceDateTime, e->cache.triggerNextEventDateTime))
    goto cleanup_and_exit;

  /* Init the Flags. */
  e->neverTriggerEvent          = false;

  e->cache.triggerCurrentEvent 	= false;

  e->cache.nextEventIsFirst 		= true;
  e->cache.lastEventWasFinal          = false;

  return e;

  cleanup_and_exit:
		if (e)
		  {
		    if (e->eventName)
		      {
			free(e->eventName);
			e->eventName = NULL;
		      }
		    deallocateDateTime(e->eventReferenceDateTime);
		    e->eventReferenceDateTime = NULL;
		    deallocateDateTime(e->eventFirstDateTime);
		    e->eventFirstDateTime = NULL;
		    deallocateDateTime(e->eventLastDateTime);
		    e->eventLastDateTime = NULL;
		    deallocateTimeDelta(e->eventInterval);
		    e->eventInterval = NULL;
                    deallocateTimeDelta(e->eventOffset);
                    e->eventOffset = NULL;
		    deallocateDateTime(e->cache.triggeredPreviousEventDateTime);
		    e->cache.triggeredPreviousEventDateTime = NULL;
                    deallocateDateTime(e->cache.triggerNextEventDateTime);
                    e->cache.triggerNextEventDateTime = NULL;
		    free(e);
		    e = NULL;
		  }
		return NULL ;
  }
  else
    return NULL;
}


/**
 * @brief Destructor of Event.
 *
 * @param  e
 *         A pointer to struct _event. e is deallocated.
 */

void
deallocateEvent(struct _event* e)
{
  if (e != NULL )
    {
      EVENTID = EVENTID - 1;

      if(e->eventName)
        {
          free(e->eventName);
          e->eventName = NULL;
       }

      if (e->cache.eventsLastEvaluationDateTime != NULL)
	deallocateDateTime(e->cache.eventsLastEvaluationDateTime);
      deallocateDateTime(e->eventReferenceDateTime);
      deallocateDateTime(e->eventFirstDateTime);
      deallocateDateTime(e->eventLastDateTime);
      deallocateDateTime(e->cache.triggerNextEventDateTime);
      deallocateDateTime(e->cache.triggeredPreviousEventDateTime);

      deallocateTimeDelta(e->eventInterval);
      deallocateTimeDelta(e->eventOffset);

      /* Warning: Do not free the NextEventInGroup. That's not your job. */
      e->nextEventInGroup = NULL;

      free(e);
      e = NULL;
    }
}


/* INTERNAL FUNCTION. */
//Get if trigger is true. Trigger true in [T-minus_slack,T+plus_slack].
static 
compare_return_val
isTriggerTimeInRange(struct _datetime* current_dt, struct _datetime* triggerNextEventDateTime, struct _timedelta* plus_slack, struct _timedelta* minus_slack)
{
  int cmp_val_flag = -128;

  /* Make a local copy of slack to avoid updating the user supplied timedeltas. */
  struct _timedelta* plus_slack_local  = NULL;
  struct _timedelta* minus_slack_local = NULL;

  if (plus_slack)
          plus_slack_local  = constructAndCopyTimeDelta(plus_slack);
  else
          plus_slack_local  = newTimeDelta(initDummyTDString); //Translates to 0.

  if (minus_slack)
          minus_slack_local = constructAndCopyTimeDelta(minus_slack);
  else
          minus_slack_local = newTimeDelta(initDummyTDString); //Translates to 0.



  /* If plus_slack_local is defined, return the status of current_dt vis-a-vis trigger time +/- allowed delta.  */
  if((plus_slack_local->sign == '+') && ((minus_slack_local->sign == '+')))
    {
      int upper_val_flag = -128;
      int lower_val_flag = -128;

      struct _datetime *dt_upperbound              = newDateTime(initDummyDTString);
      struct _datetime *dt_lowerbound              = newDateTime(initDummyDTString);

      minus_slack_local->sign = '-'; /* Change sign to obtain subtraction. */


      upper_val_flag = compareDatetime(current_dt,addTimeDeltaToDateTime(triggerNextEventDateTime,plus_slack_local,dt_upperbound));
      lower_val_flag = compareDatetime(current_dt,addTimeDeltaToDateTime(triggerNextEventDateTime,minus_slack_local,dt_lowerbound));

      if ( 
           (upper_val_flag == less_than || upper_val_flag == equal_to)
           &&
           (lower_val_flag == greater_than || lower_val_flag == equal_to)
         )
        {
          cmp_val_flag = equal_to;
        }
      else if (upper_val_flag == greater_than)
        {
          cmp_val_flag = greater_than;
        }
      else if (lower_val_flag == less_than)
        {
          cmp_val_flag = less_than;
        }

      //Cleaup.
      deallocateDateTime(dt_upperbound);
      deallocateDateTime(dt_lowerbound);
    }
  else /* If slack is malformed (negative sign is not permitted), return as normal (follow exact match for equal).  */
    {
      cmp_val_flag = compareDatetime(current_dt,triggerNextEventDateTime);
    }
   
  deallocateTimeDelta(plus_slack_local);
  deallocateTimeDelta(minus_slack_local);

  return cmp_val_flag;
}


/**
 * @brief Check if this event is active by comparing event's trigger time with current_dt.
 *
 *        The current_dt must lie in event's trigger time 
 *        (subject to optional specified slack: [Trigger_time - minus_slack, Trigger_time + plus_slack]. Slacks can be NULL. Always inclusive.)
 *        The lib has no built-in clock but relies on isCurrentEventActive(.) being called (polled) from the application at fixed intervals.
 *
 * @param  event
 *         A pointer to struct _event. This is the event being tested.
 *
 * @param  current_dt
 *         A pointer to struct _datetime. This is the 'current' datetime of the system.
 *
 * @param  plus_slack
 *         A pointer to struct _timedelta. Events are triggered between [actual_trigger_time, actual_trigger_time + plus_slack]. 
 * 	   Can be NULL; logically 0. Sign MUST be '+'
 *
 * @param  minus_slack
 *         A pointer to struct _timedelta. Events are triggered between [actual_trigger_time - minus_slack, actual_trigger_time]. 
 *         Can be NULL; logically 0. Sign MUST be '+'
 *
 * @return bool
 *         true/false indicating if the event is active.
 */

bool
isCurrentEventActive_old(struct _event* event, struct _datetime* current_dt, struct _timedelta* plus_slack, struct _timedelta* minus_slack)
{
  
  if ((event != NULL ) && (current_dt != NULL)) { // allowed slacks can be NULL.
    
    if (event->neverTriggerEvent)
      {
        return false;
      }

    /* If Event has been triggered for a datetime and checked once again return that state */
    
    if (event->cache.eventsLastEvaluationDateTime != NULL)
      {
	if (compareDatetime(current_dt, event->cache.eventsLastEvaluationDateTime) == equal_to)
	  {
	    return true;
	  }
      }
    
    /* Reset flags. */
    event->cache.triggerCurrentEvent        = false;

    //++++++++ Events trigger [start-end]. Slack is ignored outside this range. ++++++++++//
    /* If current_dt is yet not in trigger range, Return on false. */
    if (compareDatetime(current_dt,event->eventFirstDateTime) == less_than)
      return false;
    /* If last date is defined, check if current_dt is ahead of event->eventLastDateTime. Return on false. */
    if (event->eventLastDateTime && (compareDatetime(current_dt,event->eventLastDateTime) == greater_than))
      {
	/* If current_dt > event->eventLastDateTime and the event has never triggered, event->cache.lastEventWasFinal should stay false. */
	if(!event->cache.nextEventIsFirst)
	  event->cache.lastEventWasFinal = true;  //No further trigger possible.
	
	return false;
      }


    /* In case the current_dt is ahead of event->cache.triggerNextEventDateTime, we need to update the event->cache.triggerNextEventDateTime 
       to the next possible trigger time or else the events will never trigger. */
    if (isTriggerTimeInRange(current_dt, event->cache.triggerNextEventDateTime, plus_slack, NULL) == greater_than)
      {
	/* Get the first trigger time and update. */
	struct _timedelta* modulo_td = newTimeDelta(initDummyTDString);
	moduloTimeDeltaFromDateTime(event->cache.triggerNextEventDateTime, event->eventInterval, current_dt, modulo_td);
	addTimeDeltaToDateTime(current_dt,modulo_td,event->cache.triggerNextEventDateTime);
	/* Cleanup. */
	deallocateTimeDelta(modulo_td);
      }

    /* Check if trigger time is now. Trigger allowed with a slack provided [start-end] condition is met. */
    if( (!event->cache.lastEventWasFinal) && (isTriggerTimeInRange(current_dt, event->cache.triggerNextEventDateTime, plus_slack, minus_slack) == equal_to) )
      {
	/* Set previous-triggered-datetime to now. */
	replaceDatetime(current_dt, event->cache.triggeredPreviousEventDateTime);

	int cmp_val_flag = -128;
	struct _datetime* tmp_dt = newDateTime(initDummyDTString);

	/* Set the new next-trigger datetime. If eventLastDateTime is defined, then update should not happen
	   if current_dt + eventInterval > eventLastDateTime */
	if ( 
            !( event->eventLastDateTime 
               && 
               ( 
		( cmp_val_flag = isTriggerTimeInRange ( addTimeDeltaToDateTime(current_dt, event->eventInterval,tmp_dt), 
							event->eventLastDateTime, 
							NULL,
							minus_slack
							)
		  ) == greater_than 
		 )
	       )
	     ) 
	  {
	    addTimeDeltaToDateTime(event->cache.triggerNextEventDateTime,event->eventInterval,event->cache.triggerNextEventDateTime);
	  }
	else
	  event->cache.lastEventWasFinal = true;
	
	deallocateDateTime(tmp_dt); // Cleanup.
	
	/* Set event.*/
	event->cache.triggerCurrentEvent = true;      

	/* Reset indicating this is no longer true. */
	event->cache.nextEventIsFirst = false;

	if (event->cache.eventsLastEvaluationDateTime == NULL)
	  {
	    event->cache.eventsLastEvaluationDateTime = newDateTime(initDummyDTString);
	  }

	event->cache.eventsLastEvaluationDateTime = replaceDatetime(current_dt, event->cache.eventsLastEvaluationDateTime);

	return true;
      }

    return false;
  }
  else

    return false;
}

bool
isCurrentEventActive_stateless(struct _event* e, struct _datetime* query_date) 
{
  struct _datetime    *ref_date = constructAndCopyDateTime(e->eventReferenceDateTime);
  struct _datetime    *start_date = constructAndCopyDateTime(e->eventFirstDateTime);
  struct _datetime    *end_date   = constructAndCopyDateTime(e->eventLastDateTime);
  struct _julianday   *qd, *jd1, *jd2;
  struct _juliandelta *jdd;
  //#define DEBUG
    
  /* --- calculate Julian dates: */
  jd1 = newJulianDay(0, 0);
  jd2 = newJulianDay(0, 0);  
  getJulianDayFromDateTime(ref_date,   jd1);
  getJulianDayFromDateTime(end_date,   jd2);

  jdd = newJulianDelta('+', 0, 0);  
  qd  = newJulianDay(0, 0);
  getJulianDayFromDateTime(query_date, qd);

  /* --- approx. smallest and largest no. of msecs in interval */
  double ms0 = ((((e->eventInterval->day)*24.   +
		  e->eventInterval->hour)*60.   + 
		 e->eventInterval->minute)*60.  +
		e->eventInterval->second)*1000. + (double) e->eventInterval->ms;
  /* --- approx. smallest and largest no. of msecs in interval */
  double ms1 = ms0 + (e->eventInterval->year*365. + e->eventInterval->month*28.) * NO_OF_MS_IN_A_DAY;
  double ms2 = ms0 + (e->eventInterval->year*366. + e->eventInterval->month*31.) * NO_OF_MS_IN_A_DAY;

  /* split the interval into months + remainder: */
  struct _timedelta *interval1 = constructAndCopyTimeDelta(e->eventInterval);
  //struct _timedelta *interval2 = newRawTimeDelta('+', 0, e->eventInterval->month, 0, 0, 0, 0, 0);
  interval1->month = 0;

  struct _datetime  *bound_date = constructAndCopyDateTime(ref_date);  /* := ref_date + interval */
  struct _datetime  *test_date  = constructAndCopyDateTime(ref_date);
  struct _timedelta *scaled_td1 = newTimeDelta("PT0S");

  compare_return_val cmp1 = compareDatetime(query_date, end_date);
  compare_return_val cmp2 = compareDatetime(query_date, start_date);
  // we should not ask for the reference date but for the start date here:
  int done = (cmp1 == greater_than) || (cmp2 == less_than);
  int it   = 0;
  
  while (!done)
    {
#ifdef DEBUG      
      fprintf(stderr, "iteration = %d\n", it);
#endif
      it += 1;
      if (it == 5) break;
      char dt_string[MAX_DATETIME_STR_LEN];
#ifdef DEBUG
      printf("between       %s\n",   datetimeToString(ref_date, dt_string));
      printf("and           %s\n\n", datetimeToString(bound_date, dt_string));
#endif

      /* --- re-calculate Julian delta: */
      subtractJulianDay(jd2, jd1, jdd);
      double msecs_jdd = jdd->day*NO_OF_MS_IN_A_DAY + jdd->ms;      
      
      /* --- compute a datetime object mdate = ref_date + k*interval
         --- approximating 0.5*(ref_date+end_date): */
      int64_t k = (int64_t) (0.25*msecs_jdd/ms1*(ms1+ms2)/ms2);
#ifdef DEBUG
      fprintf(stderr, "k = %d\n",k);
#endif
      elementwiseScalarMultiplyTimeDelta(interval1, k, scaled_td1);
      
      addTimeDeltaToDateTime(ref_date, scaled_td1, test_date);

#ifdef DEBUG
      printf("scaled_td1    %s\n\n", timedeltaToString(scaled_td1, dt_string));
      printf("interval1    %s\n\n", timedeltaToString(interval1, dt_string));
      printf("test date     %s\n\n", datetimeToString(test_date, dt_string));
#endif
      if (k == 0 && compareDatetime(ref_date, test_date) == equal_to)
        {
          done = true;
        }
      
//      /* scalar multiply: we treat the months separately: */
//      if (interval2->month > 0)
//	{
//          int64_t k_m = k*interval2->month;
//	  struct _timedelta *scaled_td2 = newRawTimeDelta('+', k_m / 12, k_m % 12, 0, 0, 0, 0, 0);
//	  addTimeDeltaToDateTime(test_date, scaled_td2, test_date);
//	  deallocateTimeDelta(scaled_td2);
//	}

  
      /* --- select interval for next bisection step: */
      if (compareDatetime(query_date, test_date) == less_than) 
        {
          end_date = replaceDatetime(test_date, end_date);
          getJulianDayFromDateTime(end_date, jd2);
        }
      else
        {
          ref_date = replaceDatetime(test_date, ref_date);
          getJulianDayFromDateTime(ref_date, jd1);
          addTimeDeltaToDateTime(ref_date, e->eventInterval, bound_date);

          /* --- abort criterion: */
          cmp1 = compareDatetime(query_date, bound_date);
          cmp2 = compareDatetime(query_date, ref_date);

#ifdef DEBUG
	  char dt_string[MAX_DATETIME_STR_LEN];
	  printf("cmp1: %s\n",   datetimeToString(bound_date, dt_string));      
	  printf("cmp2: %s\n",   datetimeToString(ref_date, dt_string));
#endif
          done = !((cmp1 == greater_than) || (cmp2 == less_than));
        }
      
    }

#ifdef DEBUG
  char dt_string[MAX_DATETIME_STR_LEN];
  printf("after %d bisection steps:\n", it);
  printf("query date: %s\n",     datetimeToString(query_date, dt_string));      
  printf("between       %s\n",   datetimeToString(ref_date, dt_string));
  printf("and           %s\n\n", datetimeToString(bound_date, dt_string));
#endif
  
  /* --- clean up: */
  deallocateJulianDay(jd1);
  deallocateJulianDay(jd2);
  deallocateJulianDay(qd);
  deallocateJulianDelta(jdd);
  deallocateDateTime(test_date);
  deallocateDateTime(ref_date);
  deallocateDateTime(start_date);
  deallocateDateTime(end_date);
  deallocateDateTime(bound_date);
  deallocateTimeDelta(interval1);
  //  deallocateTimeDelta(interval2);
  deallocateTimeDelta(scaled_td1);

  return ((cmp1 == equal_to) || (cmp2 == equal_to));
}

bool
isCurrentEventActive(struct _event* e, struct _datetime* query_date, struct _timedelta* plus_slack, struct _timedelta* minus_slack) 
{
  if ((e != NULL ) && (query_date != NULL))
    {
#ifdef DEBUG      
      char *eString = (char *) calloc(MAX_EVENT_STR_LEN, sizeof(char));
      fprintf(stderr, "\nEvent: %s", eventToString(e,eString));
      free(eString);
#endif
      // do never, ever trigger
      if (e->neverTriggerEvent)
        {
          return false;
        }
      // date has been evaluate in previous call, so we do know the
      // result
      if (e->cache.cacheIsValid && e->cache.eventsLastEvaluationDateTime != NULL)
        {
          if (compareDatetime(query_date, e->cache.eventsLastEvaluationDateTime) == equal_to)
            {
              return true;
            }
        }      
      // reset flags
      e->cache.triggerCurrentEvent        = false;

      // event triggers [start-end]. Slack is ignored outside this
      // range

      // if query_date is not in trigger range, return false
      if (compareDatetime(query_date, e->eventFirstDateTime) == less_than)
        {
          return false;
        }
      
      // if last date is defined, check if current_dt is later than
      // event->eventLastDateTime, return false
      if (e->eventLastDateTime && (compareDatetime(query_date, e->eventLastDateTime) == greater_than))
        {
					// FIXME: Think again about the implications of the cache on this.
          // if query_date > event->eventLastDateTime and the event
          // has never triggered, event->cache.lastEventWasFinal should stay
          // false.
          if (e->cache.cacheIsValid && ! e->cache.nextEventIsFirst)
            {
              // disable future triggering
              e->cache.lastEventWasFinal = true;  
            }
          return false;
        }

      e->cache.triggerCurrentEvent = isCurrentEventActive_stateless(e, query_date);

      if (e->cache.triggerCurrentEvent)
        {
          if (e->cache.eventsLastEvaluationDateTime)
            {
              e->cache.eventsLastEvaluationDateTime = replaceDatetime(query_date, e->cache.eventsLastEvaluationDateTime);
            }
          else
            {
              e->cache.eventsLastEvaluationDateTime = constructAndCopyDateTime(query_date);
            }

					// FIXME: This again needs some thought when using the cache.
          if (e->cache.nextEventIsFirst)
            e->cache.nextEventIsFirst = false;

          replaceDatetime(query_date, e->cache.triggeredPreviousEventDateTime);

          replaceDatetime(query_date, e->cache.triggerNextEventDateTime);
          addTimeDeltaToDateTime(e->cache.triggerNextEventDateTime, e->eventInterval, e->cache.triggerNextEventDateTime);

          if (compareDatetime(e->cache.triggerNextEventDateTime, e->eventLastDateTime) == greater_than)
              e->cache.lastEventWasFinal = true;

					// Now most fields of the cache are set to useful values.
					e->cache.cacheIsValid = true;

        }
    }
  return (e->cache.triggerCurrentEvent);
}
#undef DEBUG

/* Is next event in next day(s) of previous event. */
bool
iseventNextInNextDay(struct _event* e)
{
  if( e != NULL )
    {
      bool ret = false;

      struct _date* d_next = newDate(initDummyDString);
      struct _date* d_prev = newDate(initDummyDString);
      d_next = convertDateTimeToDate(e->cache.triggerNextEventDateTime,d_next);
      d_prev = convertDateTimeToDate(e->cache.triggeredPreviousEventDateTime,d_prev);

      if(compareDate(d_next,d_prev) == greater_than)
	{
	  ret = true;
	}
      else
	{
	  ret = false;
	}

      deallocateDate(d_next);
      deallocateDate(d_prev);
      
      return ret;
    }
  else
    return false;
}


/* Is next event in next month(s) of previous event. */
bool
iseventNextInNextMonth(struct _event* e)
{
  if( e != NULL )
    {
      bool ret = false;

      struct _date* d_next = newDate(initDummyDString);
      struct _date* d_prev = newDate(initDummyDString);
      d_next = convertDateTimeToDate(e->cache.triggerNextEventDateTime,d_next);
      d_prev = convertDateTimeToDate(e->cache.triggeredPreviousEventDateTime,d_prev);

      d_next->day = 1;
      d_prev->day = 1;

      if(compareDate(d_next,d_prev) == greater_than)
	{
	  ret = true;
	}
      else
	{
	  ret = false;
	}

      deallocateDate(d_next);
      deallocateDate(d_prev);

      return ret;
    }
  else
    return false;
}

/* Is next event in next year(s) of previous event. */
bool
iseventNextInNextYear(struct _event* e)
{
  if( e != NULL )
    {
      bool ret = false;

      struct _date* d_next = newDate(initDummyDString);
      struct _date* d_prev = newDate(initDummyDString);
      d_next = convertDateTimeToDate(e->cache.triggerNextEventDateTime,d_next);
      d_prev = convertDateTimeToDate(e->cache.triggeredPreviousEventDateTime,d_prev);

      d_next->day = 1;
      d_next->month = 1;
      
      d_prev->day = 1;
      d_prev->month = 1;


      if(compareDate(d_next,d_prev) == greater_than)
	{
	  ret = true;
	}
      else
      {
	ret = false;
      }

      deallocateDate(d_next);
      deallocateDate(d_prev);

      return ret;
    }
  else
    return false;
}


/**
 * @brief Get Event as a string.
 *
 * @param  e
 *         A pointer to struct _event. The event to be converted to string.
 *
 * @param  string
 *         A pointer to char. String where event is to be written.
 *
 * @return string
 *         A pointer to the string containing event description.
 */
char*
eventToString(struct _event *e, char *eventStr)
{
  char *datetimeString = (char *) calloc(MAX_DATETIME_STR_LEN, sizeof(char));
  char *timedeltaString = (char *) calloc(MAX_TIMEDELTA_STR_LEN, sizeof(char));

  char *startStr;
  int offset;

  startStr = eventStr;
  
  offset = sprintf(startStr, "Event %s\n", e->eventName);  
  offset = sprintf(startStr+=offset, "   id                   : %" PRId64 "\n", e->eventId);
  
  if ( ! e->nextEventInGroup)
    {
      offset = sprintf(startStr+=offset, "   next group's event id: n/a\n");
    }
  else
    {
      offset = sprintf(startStr+=offset, "   next group's event id: %" PRId64 "\n", e->nextEventInGroup->eventId);      
    }
  //
  if (e->cache.eventsLastEvaluationDateTime)
    {
      datetimeString = datetimeToString(e->cache.eventsLastEvaluationDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   last evaluated       : %s\n", datetimeString); 
    }
  if (e->eventReferenceDateTime)
    {
      datetimeString = datetimeToString(e->eventReferenceDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   reference time       : %s\n", datetimeString);
    }
  //
  if (e->eventFirstDateTime)
    {
      datetimeString = datetimeToString(e->eventFirstDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   first time           : %s\n", datetimeString);  
    }
  if (e->eventLastDateTime)
    {
      datetimeString = datetimeToString(e->eventLastDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   last time            : %s\n", datetimeString);  
    }
  //
  if (e->eventInterval)
    {
      timedeltaString = timedeltaToString(e->eventInterval, timedeltaString);
      offset = sprintf(startStr+=offset, "   interval             : %s\n", timedeltaString);  
    }
  if (e->eventOffset)
    {
      timedeltaString = timedeltaToString(e->eventOffset, timedeltaString);
      offset = sprintf(startStr+=offset, "   offset               : %s\n", timedeltaString);  
    }
  //
  offset = sprintf(startStr+=offset, "   never trigger        : %s\n", e->neverTriggerEvent ? "true" : "false");  
  if (e->cache.triggerCurrentEvent)
    offset = sprintf(startStr+=offset, "   trigger              : %s\n", e->cache.triggerCurrentEvent ? "true" : "false");
  if (e->cache.nextEventIsFirst)
    offset = sprintf(startStr+=offset, "   next is first        : %s\n", e->cache.nextEventIsFirst ? "true" : "false");  
  if (e->cache.lastEventWasFinal)
    offset = sprintf(startStr+=offset, "   last was final       : %s\n", e->cache.lastEventWasFinal ? "true" : "false");

  //
  if (e->cache.triggerNextEventDateTime)
    {
      datetimeString = datetimeToString(e->cache.triggerNextEventDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   next time            : %s\n", datetimeString);  
    }
  if (e->cache.triggeredPreviousEventDateTime)
    {
      datetimeString = datetimeToString(e->cache.triggeredPreviousEventDateTime, datetimeString);
      offset = sprintf(startStr+=offset, "   previous time        : %s\n", datetimeString);  
    }
  //
  offset = sprintf(startStr+=offset, "\n");
  //
  free(datetimeString);
  free(timedeltaString);
  
  return(eventStr);
}


/* Calculate the first trigger time. First trigger is the nearest allowed trigger time >= start_dt. 
   Triggers are allowd only ref_dt + N * timestep where N can be positive, negative or 0. */
static 
struct _datetime*
getEventFirstTriggerDateTime(struct _datetime* start_dt, struct _timedelta* timestep, struct _timedelta* offset, struct _datetime* ref_dt, struct _datetime* first_trigger_dt)
{
  if ((start_dt != NULL) && (timestep != NULL) && (ref_dt != NULL) && (first_trigger_dt != NULL) ) //offset can be NULL.
    { 

      if (timestep->year != 0 && timestep->month == 0 && timestep->day == 0
          && timestep->hour == 0 && timestep->minute == 0 && timestep->second == 0 && timestep->ms == 0)
        {
          /* NOTE: years only is a trivial case. */
          
          /* Get anchor. ref + offset is the real anchor. */
          struct _datetime *anchor = constructAndCopyDateTime(ref_dt);
          addTimeDeltaToDateTime(ref_dt, offset, anchor); // Note: If offset is null, no addition takes place. 
          
	  if (compareDatetime(anchor, start_dt) == greater_than)
	    {
	      replaceDatetime(anchor, first_trigger_dt);
	    }
	  else
	    {
	      /* Determine difference between anchor and start year */
	      int64_t differenceOfYears = start_dt->date.year - anchor->date.year; // result always >= 0.
	      int64_t yearsToAdd = differenceOfYears % timestep->year;
          
	      /* We only need to update the year */
	      replaceDatetime(anchor, first_trigger_dt);
	      first_trigger_dt->date.year += yearsToAdd;
	    }
	      
          /* Cleanup. */
          deallocateDateTime(anchor);

          /* And return. */
          return first_trigger_dt;
        }
      
      else if (timestep->year == 0 && timestep->month != 0 && timestep->day == 0
               && timestep->hour == 0 && timestep->minute == 0 && timestep->second == 0 && timestep->ms == 0)
        {
          /* NOTE: months only is again a trivial case. */

          /* Get anchor. ref + offset is the real anchor. */
          struct _datetime *anchor = constructAndCopyDateTime(ref_dt);
          addTimeDeltaToDateTime(ref_dt, offset, anchor); // Note: If offset is null, no addition takes place. 

	  if (compareDatetime(anchor, start_dt) == greater_than)
	    {
	      replaceDatetime(anchor, first_trigger_dt);
	    }
	  else
	    {
	      /* Determine difference between anchor and start in months */
	      int differenceOfMonths = 12*(start_dt->date.year - anchor->date.year) + start_dt->date.month - anchor->date.month;

	      int yearsToAdd = differenceOfMonths / 12;
	      int monthsToAdd = differenceOfMonths % timestep->month;

	      /* We only need to update the year and month */
	      replaceDatetime(anchor, first_trigger_dt);
	      first_trigger_dt->date.year += yearsToAdd;
	      first_trigger_dt->date.month += monthsToAdd;
	    }
	      
          /* Cleanup. */
          deallocateDateTime(anchor);

          /* And return. */
          return first_trigger_dt;
        }
      
      else if (timestep->year == 0 && timestep->month == 0)
        
        {
          /* NOTE: This code is higly 'rigged', to a point where it might feel micromanaged.   
             This is to speed-up the event-init process. Without the hack, the code
             will be too slow and hence the ends justify the means. 
          */

          struct _timedelta* timestep_bkw;
          
          /* Get start date in julian. */
          struct _julianday* start_jd = newJulianDay(0, 0);
          start_jd = date2julian(start_dt, start_jd);
          
          
          /* Get timedelta in juliandelta. */
          struct _juliandelta* timestep_jd = newJulianDelta('+', 0, 0);
          timestep_jd = timeDeltaToJulianDelta(timestep, ref_dt, timestep_jd);
          
          
          /* For speed-up */
          struct _juliandelta* timestep_fastfwd_jd = newJulianDelta('+', 0, 0);
          struct _juliandelta* timestep_fastbkw_jd = newJulianDelta('+', 0, 0);
          
          
          /* Get anchor. ref + offset is the real anchor. */
          struct _datetime* anchor = constructAndCopyDateTime(ref_dt);
          addTimeDeltaToDateTime(ref_dt, offset, anchor); // Note: If offset is null, no addition takes place. 
          
          /* Get anchor in julian. */
          struct _julianday* anchor_jd = newJulianDay(0, 0);
          anchor_jd = date2julian(anchor, anchor_jd);
          
          /* Optimization hack: Calculate an approx metric are_dates_too_far and speed up the jumps.  */
          int64_t are_dates_too_far = 0;
          if (timestep_jd->day)
            are_dates_too_far = (start_jd->day - anchor_jd->day)/(timestep_jd->day); 
          else if (timestep_jd->ms)
            are_dates_too_far = (start_jd->day - anchor_jd->day)/((float)timestep_jd->ms/NO_OF_MS_IN_A_DAY);
          //else ... well, should never happen. If it does, the initialized value of zero persists.
          
          
          int lambda = 1; //speed-up const. Default is 1 or no speedup.
          if (are_dates_too_far>10 || are_dates_too_far<-10)
            lambda = 100000; //speed up if start-date and anchor are 'too-far' away. 
          
          
          /* Fast-Fwd */
          timestep_fastfwd_jd->day = lambda*timestep_jd->day;
          timestep_fastfwd_jd->ms  = lambda*timestep_jd->ms;
          
          while( timestep_fastfwd_jd->ms >= NO_OF_MS_IN_A_DAY )
            {
              timestep_fastfwd_jd->day++;
              timestep_fastfwd_jd->ms = timestep_fastfwd_jd->ms - NO_OF_MS_IN_A_DAY;
            }
          
          
          /* We need to Loop backwards: Create a timestep replica and change the sign to negative to travel-back-in-time. */
          timestep_bkw = constructAndCopyTimeDelta(timestep);
          timestep_bkw->sign = '-';
          
          struct _juliandelta* timestep_bkw_jd = newJulianDelta('+', 0, 0);
          timeDeltaToJulianDelta(timestep_bkw, ref_dt, timestep_bkw_jd);
          
          /* Fast-Bkwd */
          timestep_fastbkw_jd->day = lambda*timestep_bkw_jd->day;
          timestep_fastbkw_jd->ms  = lambda*timestep_bkw_jd->ms;
          
          while( timestep_fastbkw_jd->ms <=  (-1)*(NO_OF_MS_IN_A_DAY) )
            {
              timestep_fastbkw_jd->day--;
              timestep_fastbkw_jd->ms = timestep_fastbkw_jd->ms + NO_OF_MS_IN_A_DAY;
            }
          
          
          switch (compareDatetime(start_dt,ref_dt))
            { 
            case greater_than: /* start_dt > ref_dt */      
              
              /* Jump very fast and reach the start date quickly. */
              do
                {
                  anchor_jd->day = anchor_jd->day + timestep_fastfwd_jd->day;
                  anchor_jd->ms  = anchor_jd->ms  + timestep_fastfwd_jd->ms;
                  
                  if (anchor_jd->ms >= NO_OF_MS_IN_A_DAY)
                    {
                      anchor_jd->day = anchor_jd->day + 1;
                      anchor_jd->ms = anchor_jd->ms - NO_OF_MS_IN_A_DAY;
                    }
                  
                } while ( !((anchor_jd->day > start_jd->day)  ||  (anchor_jd->day == start_jd->day && anchor_jd->ms > start_jd->ms)) );
              
              
              /* But I jumped one time too much. Move back */
              anchor_jd->day = anchor_jd->day - timestep_fastfwd_jd->day;
              anchor_jd->ms  = anchor_jd->ms  - timestep_fastfwd_jd->ms;
              
              if (anchor_jd->ms < 0)
                {
                  anchor_jd->day = anchor_jd->day - 1;
                  anchor_jd->ms = anchor_jd->ms + NO_OF_MS_IN_A_DAY;
                }
              
              
              
              /* I am close. Now determine the actual time. */
              do
                {
                  anchor_jd->day = anchor_jd->day + timestep_jd->day;
                  anchor_jd->ms = anchor_jd->ms + timestep_jd->ms;
                  
                  if (anchor_jd->ms >= NO_OF_MS_IN_A_DAY)
                    {
                      anchor_jd->day = anchor_jd->day + 1;
                      anchor_jd->ms = anchor_jd->ms - NO_OF_MS_IN_A_DAY;
                    }
                  
                } while ( !((anchor_jd->day > start_jd->day)  ||  (anchor_jd->day == start_jd->day && anchor_jd->ms >= start_jd->ms)) );
              
              /* anchor_jd is now the true event-trigger-time. Set it to anchor. */
              julian2date(anchor_jd,anchor);
              
              break;
              
            case equal_to: /* start_dt == ref_dt */
              
              break;
              
            case less_than: /* start_dt < ref_dt */
             
              
              /* Jump very fast bkwd and reach the start date quickly. */
              do
                {
                  anchor_jd->day = anchor_jd->day + timestep_fastbkw_jd->day;
                  anchor_jd->ms = anchor_jd->ms + timestep_fastbkw_jd->ms;
                  
                  if ( anchor_jd->ms < 0 )
                    {
                      anchor_jd->day = anchor_jd->day - 1;
                      anchor_jd->ms = anchor_jd->ms + NO_OF_MS_IN_A_DAY;
                    }
                  
                } while(! (anchor_jd->day < start_jd->day  ||  (anchor_jd->day == start_jd->day && anchor_jd->ms < start_jd->ms)) );
              
              
              /* I jumped one time too much. Move forward. */
              anchor_jd->day = anchor_jd->day + timestep_fastfwd_jd->day;
              anchor_jd->ms = anchor_jd->ms + timestep_fastfwd_jd->ms;
              
              if (anchor_jd->ms >= NO_OF_MS_IN_A_DAY)
                {
                  anchor_jd->day = anchor_jd->day + 1;
                  anchor_jd->ms = anchor_jd->ms - NO_OF_MS_IN_A_DAY;
                }
              
              
              /* I am close. Get the real time. */
              do
                {
                  anchor_jd->day = anchor_jd->day + timestep_bkw_jd->day;
                  anchor_jd->ms = anchor_jd->ms + timestep_bkw_jd->ms;
                  
                  if ( anchor_jd->ms < 0 )
                    {
                      anchor_jd->day = anchor_jd->day - 1;
                      anchor_jd->ms = anchor_jd->ms + NO_OF_MS_IN_A_DAY;
                    }
                  
                } while(! (anchor_jd->day < start_jd->day  ||  (anchor_jd->day == start_jd->day && anchor_jd->ms < start_jd->ms)) );
              
              
              /* I jumped one time too much. Move forward.  */
              anchor_jd->day = anchor_jd->day + timestep_jd->day;
              anchor_jd->ms = anchor_jd->ms + timestep_jd->ms;
              
              if (anchor_jd->ms >= NO_OF_MS_IN_A_DAY)
                {
                  anchor_jd->day = anchor_jd->day + 1;
                  anchor_jd->ms = anchor_jd->ms - NO_OF_MS_IN_A_DAY;
                }
              
              /* anchor_jd now has the true trigger time. Copy it to anchor. */
              julian2date(anchor_jd,anchor);
              
            default:
              
              ; //Should never happen. Don't return NULL just yet as deallocateDateTime(anchor) still needs to be called.
              
            }
          
          /* Copy the contents to target. */
          replaceDatetime(anchor,first_trigger_dt);
          
          
          /* Cleanup. */
          
          deallocateDateTime(anchor);
          deallocateJulianDay(start_jd);
          deallocateJulianDay(anchor_jd);
          deallocateJulianDelta(timestep_jd);
          deallocateTimeDelta(timestep_bkw);
          deallocateJulianDelta(timestep_bkw_jd);
          
          deallocateJulianDelta(timestep_fastfwd_jd);
          deallocateJulianDelta(timestep_fastbkw_jd);
          
          /* And return. */
          return first_trigger_dt;
        }
      else
        {
          /* NOTE: general case updates by explicit adding. */

          /* Get anchor. ref + offset is the real anchor. */
          struct _datetime *anchor = constructAndCopyDateTime(ref_dt);
          addTimeDeltaToDateTime(ref_dt, offset, anchor); // Note: If offset is null, no addition takes place. 

          while (compareDatetime(anchor, start_dt) == less_than)
            {
              addTimeDeltaToDateTime(anchor, timestep, anchor);
            } 
          /* Copy the contents to target. */
          replaceDatetime(anchor,first_trigger_dt);
          
          /* Cleanup. */
          deallocateDateTime(anchor);

          /* And return. */
          return first_trigger_dt;          
        }
    }
  else
    return NULL;
}


/**
  * @brief Get the Datetime when 'this' event will be triggered next.
  *
  * WARNING: The value returned is with-respect-to current_dt and not a true copy of triggerNextEventDateTime in the event data structure.
  *
  * @param  e
  *         A pointer to struct _event. This is the event being queried.
  *
  * @param  dt_return
  *         A pointer to struct _datetime. The next trigger datetime is copied here.
  *
  * @return dt_return
  *         A pointer to DateTime with next-trigger datetime.
  */

struct _datetime*
getTriggerNextEventAtDateTime(struct _event* e, struct _datetime* current_dt, struct _datetime* dt_return)
{
  if ((e != NULL) && (current_dt != NULL) && (dt_return != NULL))
    {
       /* If last date is defined, check if current_dt is ahead of event->eventLastDateTime. 
          Return on NULL indicating no future trigger event scheduled.*/
      if (e->eventLastDateTime && (compareDatetime(current_dt,e->eventLastDateTime) == greater_than))
        return NULL;

      /* This if should evaluate to false normally. The only occasion when it does not is if the 
         current_dt > triggerNextEventDateTime and isCurrentEventActive(.) has never been called  */
      if (
           (e->cache.nextEventIsFirst)
           &&
           (compareDatetime(current_dt,e->cache.triggerNextEventDateTime) == greater_than)
         )
        {
              struct _timedelta* modulo_td = newTimeDelta("PT00.000S");
              // Get the first trigger time and return (WARNING: Do not update e->cache.triggerNextEventDateTime here!).
              moduloTimeDeltaFromDateTime(e->cache.triggerNextEventDateTime, e->eventInterval, current_dt, modulo_td);
              addTimeDeltaToDateTime(current_dt,modulo_td,dt_return);
              // Cleanup. 
              deallocateTimeDelta(modulo_td);
              return dt_return;
        }
      // else ...
      dt_return = replaceDatetime(e->cache.triggerNextEventDateTime, dt_return);
      return dt_return;
    } 
  else
    return NULL;
}

/**
  * @brief Get the Datetime when 'this' event was triggered last.
  *
  * @param  e
  *         A pointer to struct _event. This is the event being queried.
  *
  * @param  dt_return
  *         A pointer to struct _datetime. The last trigger datetime is copied here.
  *
  * @return dt_return
  *         A pointer to DT with previous-trigger datetime.
  */
/* NOTE: If the event was never tiggered, default value of 0-01-01T00:00:00.000 is returned. */
struct _datetime*
getTriggeredPreviousEventAtDateTime(struct _event* e, struct _datetime* dt_return)
{
  if ((e != NULL )&& (dt_return != NULL) )
    {
      /* No trigger ever happened? */
      if (e->cache.nextEventIsFirst)
        return NULL; 

      replaceDatetime(e->cache.triggeredPreviousEventDateTime, dt_return);
      return dt_return;
    }
  else
    return NULL;
}


struct _event*
getNextEventInGroup(struct _event* e)
{
  if (e != NULL)
    {
      fprintf(stderr, "lk: next event in group defined ...\n");
      return e->nextEventInGroup;   
    }
  else
    {
      fprintf(stderr, "lk: next event in group NOT defined ...\n");      
      return NULL;
    }
}

int64_t
getEventId(struct _event* e)
{
  if (e != NULL)
    return e->eventId;
  else
    return 0;
}

char*
getEventName(struct _event* e, char* ename)
{
  if ((e != NULL) && (ename != NULL))
    {
      strncpy(ename, e->eventName, MAX_EVENTNAME_STR_LEN-1);
      ename[MAX_EVENTNAME_STR_LEN-1] = '\0';
      return ename;
    }
  else
    return NULL;
}

struct _datetime*
getEventReferenceDateTime(struct _event* e)
{
  if (e != NULL)
    return e->eventReferenceDateTime;
  else
    return NULL;
}

struct _datetime*
getEventFirstDateTime(struct _event* e)
{
  if (e != NULL)
    return e->eventFirstDateTime;
  else
    return NULL;
}

struct _datetime*
getEventLastDateTime(struct _event* e)
{
  if (e != NULL)
    return e->eventLastDateTime;
  else
    return NULL;
}

struct _timedelta*
getEventInterval(struct _event* e)
{
  if (e != NULL)
    return e->eventInterval;
  else
    return NULL;
}

bool
getNextEventIsFirst(struct _event* e)
{
  if (e != NULL)
    return e->cache.nextEventIsFirst;
  else
    return false;
}
